@@ -6,12 +6,13 @@ from mohou.encoder import ImageEncoder, VectorIdenticalEncoder
 from mohou.encoding_rule import EncodingRule
 from mohou.model import LSTM, AutoEncoderBase, Chimera
 from mohou.propagator import Propagator
+from mohou.setting import setting
 from mohou.trainer import TrainCache
 from mohou.types import (
     AngleVector,
+    EpisodeBundle,
     GripperState,
     ImageBase,
-    MultiEpisodeChunk,
     TerminateFlag,
     get_all_concrete_leaftypes,
 )
@@ -22,6 +23,9 @@ class DefaultNotFoundError(Exception):
 
 
 def auto_detect_autoencoder_type(project_name: Optional[str] = None) -> Type[AutoEncoderBase]:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
 
     # TODO(HiroIshida) dirty...
     t: Optional[Type[AutoEncoderBase]] = None
@@ -47,6 +51,10 @@ def auto_detect_autoencoder_type(project_name: Optional[str] = None) -> Type[Aut
 
 
 def load_default_image_encoder(project_name: Optional[str] = None) -> ImageEncoder:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
+
     ae_type = auto_detect_autoencoder_type(project_name)
     try:
         tcache_autoencoder = TrainCache.load(project_name, ae_type)
@@ -59,29 +67,36 @@ def load_default_image_encoder(project_name: Optional[str] = None) -> ImageEncod
 
 
 def create_default_encoding_rule(project_name: Optional[str] = None) -> EncodingRule:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
 
-    chunk = MultiEpisodeChunk.load(project_name)
-    chunk_spec = chunk.spec
-    av_dim = chunk_spec.type_shape_table[AngleVector][0]
+    bundle = EpisodeBundle.load(project_name)
+    bundle_spec = bundle.spec
+    av_dim = bundle_spec.type_shape_table[AngleVector][0]
     image_encoder = load_default_image_encoder(project_name)
     av_idendical_encoder = VectorIdenticalEncoder(AngleVector, av_dim)
 
     encoders = [image_encoder, av_idendical_encoder]
 
-    if GripperState in chunk_spec.type_shape_table:
+    if GripperState in bundle_spec.type_shape_table:
         gs_identital_func = VectorIdenticalEncoder(
-            GripperState, chunk_spec.type_shape_table[GripperState][0]
+            GripperState, bundle_spec.type_shape_table[GripperState][0]
         )
         encoders.append(gs_identital_func)
 
     tf_identical_func = VectorIdenticalEncoder(TerminateFlag, 1)
     encoders.append(tf_identical_func)
 
-    encoding_rule = EncodingRule.from_encoders(encoders, chunk)
+    encoding_rule = EncodingRule.from_encoders(encoders, bundle)
     return encoding_rule
 
 
 def create_chimera_encoding_rule(project_name: Optional[str] = None) -> EncodingRule:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
+
     # experimental
     encoding_rule = create_default_encoding_rule(project_name)
     image_type = [k for k in encoding_rule.keys() if issubclass(k, ImageBase)].pop()
@@ -99,6 +114,10 @@ def create_chimera_encoding_rule(project_name: Optional[str] = None) -> Encoding
 
 
 def create_default_propagator(project_name: Optional[str] = None) -> Propagator:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
+
     try:
         tcach_lstm = TrainCache.load(project_name, LSTM)
     except Exception:
@@ -111,6 +130,10 @@ def create_default_propagator(project_name: Optional[str] = None) -> Propagator:
 
 
 def create_chimera_propagator(project_name: Optional[str] = None) -> Propagator:
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
+
     encoding_rule = create_chimera_encoding_rule(project_name)
 
     chimera = TrainCache.load(project_name, Chimera).best_model
@@ -121,14 +144,14 @@ def create_chimera_propagator(project_name: Optional[str] = None) -> Propagator:
 
 
 def create_default_image_context_list(
-    project_name: Optional[str] = None, chunk: Optional[MultiEpisodeChunk] = None
+    project_name: Optional[str] = None, bundle: Optional[EpisodeBundle] = None
 ) -> List[np.ndarray]:
-    if chunk is None:
-        chunk = MultiEpisodeChunk.load(project_name)
+    if bundle is None:
+        bundle = EpisodeBundle.load(project_name)
     image_encoder = load_default_image_encoder(project_name)
 
     context_list = []
-    for episode in chunk:
+    for episode in bundle:
         seq = episode.get_sequence_by_type(image_encoder.elem_type)
         context = image_encoder.forward(seq[0])
         context_list.append(context)
