@@ -7,7 +7,7 @@ from mohou.model.autoencoder import AutoEncoderConfig
 from mohou.script_utils import create_default_logger, train_autoencoder
 from mohou.setting import setting
 from mohou.trainer import TrainConfig
-from mohou.types import ImageBase, MultiEpisodeChunk, RGBImage, get_element_type
+from mohou.types import EpisodeBundle, ImageBase, RGBImage, get_element_type
 
 if __name__ == "__main__":
     parser = argparse.ArgumentParser()
@@ -16,7 +16,7 @@ if __name__ == "__main__":
     parser.add_argument("-aug", type=int, default=2, help="number of augmentation X")
     parser.add_argument("-latent", type=int, default=16, help="latent space dim")
     parser.add_argument("-image", type=str, default="RGBImage", help="image type")
-    parser.add_argument("-chunk_postfix", type=str, default="", help="postfix for chunk")
+    parser.add_argument("-bundle_postfix", type=str, default="", help="postfix for bundle")
     parser.add_argument(
         "-valid-ratio", type=float, default=0.1, help="split rate for validation dataset"
     )
@@ -31,16 +31,16 @@ if __name__ == "__main__":
     valid_ratio = args.valid_ratio
     use_vae = args.vae
     warm_start = args.warm
-    chunk_postfix = args.chunk_postfix
+    bundle_postfix = args.bundle_postfix
 
     logger = create_default_logger(project_name, "autoencoder")
 
-    if chunk_postfix == "":
-        chunk_postfix = None
-    chunk = MultiEpisodeChunk.load(project_name, chunk_postfix)
+    if bundle_postfix == "":
+        bundle_postfix = None
+    bundle = EpisodeBundle.load(project_name, bundle_postfix)
 
     image_type: Type[ImageBase] = get_element_type(args.image)  # type: ignore
-    n_pixel, _, _ = chunk.spec.type_shape_table[RGBImage]  # Assuming chunk contains rgb
+    n_pixel, _, _ = bundle.spec.type_shape_table[RGBImage]  # Assuming bundle contains rgb
     model_config = AutoEncoderConfig(image_type, n_bottleneck, n_pixel)
     dataset_config = AutoEncoderDatasetConfig(n_aug)
     train_config = TrainConfig(n_epoch=n_epoch, valid_data_ratio=valid_ratio)
@@ -51,7 +51,7 @@ if __name__ == "__main__":
         model_config,
         dataset_config,
         train_config,
-        chunk=chunk,
+        bundle=bundle,
         ae_type=ae_type,
         warm_start=warm_start,
     )
