@@ -8,7 +8,7 @@ from mohou.model.markov import MarkoveModelConfig
 from mohou.script_utils import create_default_logger
 from mohou.setting import setting
 from mohou.trainer import TrainCache, TrainConfig, train
-from mohou.types import AngleVector, MultiEpisodeChunk
+from mohou.types import AngleVector, EpisodeBundle
 
 
 def create_obs_rule():
@@ -32,21 +32,20 @@ if __name__ == "__main__":
 
     logger = create_default_logger(project_name, "control_model")
 
-    chunk = MultiEpisodeChunk.load(project_name)
-    n_av_dim = chunk.spec.type_shape_table[AngleVector][0]
+    bundle = EpisodeBundle.load(project_name)
+    n_av_dim = bundle.spec.type_shape_table[AngleVector][0]
     f = VectorIdenticalEncoder(AngleVector, n_av_dim)
     ctrl_rule = EncodingRule.from_encoders([f])
 
     obs_rule = create_obs_rule()
 
-    dataset = MarkovControlSystemDataset.from_chunk(
-        chunk, ctrl_rule, obs_rule, diff_as_control=True
+    dataset = MarkovControlSystemDataset.from_bundle(
+        bundle, ctrl_rule, obs_rule, diff_as_control=True
     )
 
-    tcache = TrainCache[ControlModel](project_name)
     n_input = ctrl_rule.dimension + obs_rule.dimension
     n_output = obs_rule.dimension
     config = MarkoveModelConfig(n_input, n_output, activation="relu")
-    model = ControlModel(config)
+    tcache = TrainCache[ControlModel].from_model(ControlModel(config))
     tconfig = TrainConfig(n_epoch=n_epoch)
-    train(tcache, dataset, model, config=tconfig)
+    train(project_name, tcache, dataset, config=tconfig)
