@@ -8,7 +8,7 @@ from mohou.script_utils import (
 )
 from mohou.setting import setting
 from mohou.trainer import TrainCache
-from mohou.types import MultiEpisodeChunk
+from mohou.types import EpisodeBundle
 
 if __name__ == "__main__":
     parser = argparse.ArgumentParser()
@@ -19,17 +19,17 @@ if __name__ == "__main__":
     project_name = args.pn
     n_vis = args.n
 
-    chunk = MultiEpisodeChunk.load(project_name)
+    bundle = EpisodeBundle.load(project_name)
 
     if args.chimera:
         chimera = TrainCache.load(project_name, Chimera).best_model
         assert chimera is not None
-        visualize_image_reconstruction(project_name, chunk, chimera.ae, n_vis, prefix="chimera")
+        visualize_image_reconstruction(project_name, bundle, chimera.ae, n_vis, prefix="chimera")
     else:
         ae_type = auto_detect_autoencoder_type(project_name)
         model = TrainCache.load(project_name, ae_type).best_model
         assert model is not None
-        visualize_image_reconstruction(project_name, chunk, model, n_vis)
+        visualize_image_reconstruction(project_name, bundle, model, n_vis)
 
         if ae_type == VariationalAutoEncoder:  # TODO(HiroIshida): enable this also for chimera
             visualize_variational_autoencoder(project_name)
