@@ -1,7 +1,5 @@
 import logging
-import pickle
 import random
-import re
 import time
 from logging import Logger
 from typing import List, Optional, Type, Union
@@ -35,13 +33,14 @@ from mohou.model import (
 )
 from mohou.model.chimera import Chimera, ChimeraConfig, ChimeraDataset
 from mohou.propagator import Propagator
+from mohou.setting import setting
 from mohou.trainer import TrainCache, TrainConfig, train
 from mohou.types import (
     AngleVector,
     ElementDict,
+    EpisodeBundle,
     GripperState,
     ImageBase,
-    MultiEpisodeChunk,
     TerminateFlag,
 )
 from mohou.utils import canvas_to_ndarray
@@ -77,22 +76,22 @@ def train_autoencoder(
     dataset_config: AutoEncoderDatasetConfig,
     train_config: TrainConfig,
     ae_type: Type[AutoEncoderBase] = AutoEncoder,
-    chunk: Optional[MultiEpisodeChunk] = None,
+    bundle: Optional[EpisodeBundle] = None,
     warm_start: bool = False,
 ):
 
-    if chunk is None:
-        chunk = MultiEpisodeChunk.load(project_name)
+    if bundle is None:
+        bundle = EpisodeBundle.load(project_name)
 
-    dataset = AutoEncoderDataset.from_chunk(chunk, image_type, dataset_config)
+    dataset = AutoEncoderDataset.from_bundle(bundle, image_type, dataset_config)
     if warm_start:
         logger.info("warm start")
         tcache = TrainCache.load(project_name, ae_type)
-        train(tcache, dataset, model=None, config=train_config)
+        train(project_name, tcache, dataset, config=train_config)
     else:
-        tcache = TrainCache(project_name)  # type: ignore[var-annotated]
         model = ae_type(model_config)  # type: ignore
-        train(tcache, dataset, model=model, config=train_config)
+        tcache = TrainCache.from_model(model)  # type: ignore[var-annotated]
+        train(project_name, tcache, dataset, config=train_config)
 
 
 def train_lstm(
@@ -102,13 +101,13 @@ def train_lstm(
     dataset_config: AutoRegressiveDatasetConfig,
     train_config: TrainConfig,
     weighting: Optional[Union[WeightPolicy, List[np.ndarray]]] = None,
-    chunk: Optional[MultiEpisodeChunk] = None,
+    bundle: Optional[EpisodeBundle] = None,
     warm_start: bool = False,
     context_list: Optional[List[np.ndarray]] = None,
 ):
 
-    if chunk is None:
-        chunk = MultiEpisodeChunk.load(project_name)
+    if bundle is None:
+        bundle = EpisodeBundle.load(project_name)
 
     if context_list is None:
         assert model_config.n_static_context == 0
@@ -116,8 +115,8 @@ def train_lstm(
         for context in context_list:
             assert len(context) == model_config.n_static_context
 
-    dataset = AutoRegressiveDataset.from_chunk(
-        chunk,
+    dataset = AutoRegressiveDataset.from_bundle(
+        bundle,
         encoding_rule,
         augconfig=dataset_config,
         weighting=weighting,
@@ -127,11 +126,11 @@ def train_lstm(
     if warm_start:
         logger.info("warm start")
         tcache = TrainCache.load(project_name, LSTM)
-        train(tcache, dataset, model=None, config=train_config)
+        train(project_name, tcache, dataset, config=train_config)
     else:
-        tcache = TrainCache(project_name)  # type: ignore[var-annotated]
         model = LSTM(model_config)
-        train(tcache, dataset, model=model, config=train_config)
+        tcache = TrainCache.from_model(model)  # type: ignore[var-annotated]
+        train(project_name, tcache, dataset, config=train_config)
 
 
 def train_chimera(
@@ -139,57 +138,54 @@ def train_chimera(
     encoding_rule: EncodingRule,
     lstm_config: LSTMConfig,
     train_config: TrainConfig,
-    chunk: Optional[MultiEpisodeChunk] = None,
+    bundle: Optional[EpisodeBundle] = None,
 ):  # TODO(HiroIshida): minimal args
 
-    if chunk is None:
-        chunk = MultiEpisodeChunk.load(project_name)
+    if bundle is None:
+        bundle = EpisodeBundle.load(project_name)
 
-    dataset = ChimeraDataset.from_chunk(chunk, encoding_rule)
-    tcache = TrainCache(project_name)  # type: ignore[var-annotated]
+    dataset = ChimeraDataset.from_bundle(bundle, encoding_rule)
     ae = TrainCache.load(project_name, AutoEncoder).best_model
     assert ae is not None
     conf = ChimeraConfig(lstm_config, ae_config=ae)
     model = Chimera(conf)  # type: ignore[var-annotated]
-    train(tcache, dataset, model, train_config)
+    tcache = TrainCache.from_model(model)  # type: ignore[var-annotated]
+    train(project_name, tcache, dataset, train_config)
 
 
 def visualize_train_histories(project_name: str):
     project_path = get_project_path(project_name)
-    file_paths = sorted(project_path.iterdir())
+    sorted(project_path.iterdir())
 
     plot_dir_path = get_subproject_path(project_name, "train_history")
-    for path in file_paths:
-        m = re.match(r".*TrainCache.*", path.name)
-        if m is not None:
-            pickle_path = project_path / path.name
 
-            with pickle_path.open(mode="rb") as f:
-                tcache: TrainCache = pickle.load(f)
-                fig, ax = plt.subplots()
-                tcache.visualize((fig, ax))
-                image_path = plot_dir_path / (path.name + ".png")
-                fig.savefig(str(image_path))
-                print("saved to {}".format(image_path))
+    all_result_paths = TrainCache.filter_result_paths(project_name, None, None)
+    for result_path in all_result_paths:
+        tcache = TrainCache.load_from_base_path(result_path)
+        image_path = plot_dir_path / (result_path.name + ".png")
+        fig, ax = plt.subplots()
+        tcache.visualize((fig, ax))
+        fig.savefig(str(image_path))
+        print("saved to {}".format(image_path))
 
 
 def visualize_image_reconstruction(
     project_name: str,
-    chunk: MultiEpisodeChunk,
+    bundle: EpisodeBundle,
     autoencoder: AutoEncoderBase,
     n_vis: int = 5,
     prefix: Optional[str] = None,
 ):
 
-    chunk_intact = chunk.get_intact_chunk()
-    chunk_not_intact = chunk.get_not_intact_chunk()
+    bundle_untouch = bundle.get_untouch_bundle()
+    bundle_touch = bundle.get_touch_bundle()
 
     image_type = autoencoder.image_type  # type: ignore[union-attr]
     no_aug = AutoEncoderDatasetConfig(0)  # to feed not randomized image
-    dataset_intact = AutoEncoderDataset.from_chunk(chunk_intact, image_type, no_aug)
-    dataset_not_intact = AutoEncoderDataset.from_chunk(chunk_not_intact, image_type, no_aug)
+    dataset_untouch = AutoEncoderDataset.from_bundle(bundle_untouch, image_type, no_aug)
+    dataset_touch = AutoEncoderDataset.from_bundle(bundle_touch, image_type, no_aug)
 
-    for dataset, postfix in zip([dataset_intact, dataset_not_intact], ["intact", "not_intact"]):
+    for dataset, postfix in zip([dataset_untouch, dataset_touch], ["untouch", "touch"]):
         idxes = list(range(len(dataset)))
         random.shuffle(idxes)
         idxes_test = idxes[: min(n_vis, len(dataset))]
@@ -217,6 +213,10 @@ def visualize_image_reconstruction(
 
 
 def visualize_variational_autoencoder(project_name: Optional[str] = None):
+    if project_name is None:
+        assert setting.primary_project_name is not None
+        project_name = setting.primary_project_name
+
     tcache = TrainCache.load(project_name, VariationalAutoEncoder)
     vae = tcache.best_model
     assert vae is not None
@@ -245,12 +245,12 @@ def add_text_to_image(image: ImageBase, text: str, color: str):
 
 def visualize_lstm_propagation(project_name: str, propagator: Propagator, n_prop: int):
 
-    chunk = MultiEpisodeChunk.load(project_name).get_intact_chunk()
+    bundle = EpisodeBundle.load(project_name).get_untouch_bundle()
     save_dir_path = get_subproject_path(project_name, "lstm_result")
     image_encoder = load_default_image_encoder(project_name)
 
-    for idx, edata in enumerate(chunk):
-        episode_data = chunk[idx]
+    for idx, edata in enumerate(bundle):
+        episode_data = bundle[idx]
 
         image_type = None
         for key, encoder in propagator.encoding_rule.items():
@@ -287,8 +287,8 @@ def visualize_lstm_propagation(project_name: str, propagator: Propagator, n_prop
         if use_gripper:
             pred_gss = [elem_dict[GripperState].numpy() for elem_dict in elem_dict_list]
 
-        n_av_dim = chunk.spec.type_shape_table[AngleVector][0]
-        n_gs_dim = chunk.spec.type_shape_table[GripperState][0] if use_gripper else 0
+        n_av_dim = bundle.spec.type_shape_table[AngleVector][0]
+        n_gs_dim = bundle.spec.type_shape_table[GripperState][0] if use_gripper else 0
         fig, axs = plt.subplots(n_av_dim + n_gs_dim, 1)
 
         # plot angle vectors
