@@ -1,9 +1,17 @@
 import copy
 import functools
 import hashlib
+import json
+import logging
 import operator
+import os
+import pathlib
 import pickle
 import random
+import re
+import shutil
+import subprocess
+import tempfile
 from abc import ABC, abstractmethod
 from dataclasses import asdict, dataclass
 from pathlib import Path
@@ -24,6 +32,7 @@ from typing import (
 
 import cv2
 import matplotlib.pyplot as plt
+import natsort
 import numpy as np
 import PIL.Image
 import torch
@@ -31,7 +40,7 @@ import torchvision
 import yaml
 
 from mohou.constant import CONTINUE_FLAG_VALUE, TERMINATE_FLAG_VALUE
-from mohou.file import dump_object, get_project_path, load_object
+from mohou.file import get_project_path
 from mohou.image_randomizer import (
     _f_randomize_depth_image,
     _f_randomize_gray_image,
@@ -46,6 +55,8 @@ from mohou.utils import (
     split_sequence,
 )
 
+logger = logging.getLogger(__name__)
+
 ElementT = TypeVar("ElementT", bound="ElementBase")
 PrimitiveElementT = TypeVar("PrimitiveElementT", bound="PrimitiveElementBase")
 PrimitiveImageT = TypeVar("PrimitiveImageT", bound="PrimitiveImageBase")
@@ -476,45 +487,72 @@ def get_element_type(type_name: str) -> Type[ElementBase]:
     assert False, "type {} not found".format(type_name)
 
 
+@dataclass(frozen=True)
 class ElementSequence(HasAList[ElementT], Generic[ElementT]):
-    elem_type: Optional[Type[ElementT]] = None
-    elem_shape: Optional[Tuple[int, ...]] = None
     elem_list: List[ElementT]
 
-    def __init__(self, elem_list: Optional[List[ElementT]] = None):
-        if elem_list is None or len(elem_list) == 0:
-            self.elem_list = []
-        else:
-            assert len(set([type(elem) for elem in elem_list])) == 1
-            assert len(set([elem.shape for elem in elem_list])) == 1
-            self.elem_list = elem_list
-            self.elem_type = type(elem_list[0])
-            self.elem_shape = elem_list[0].shape
+    def __post_init__(self):
+        # validation
+        assert isinstance(self.elem_list, list)
+        assert len(self.elem_list) > 0
+        assert len(set([type(elem) for elem in self.elem_list])) == 1
+        assert len(set([elem.shape for elem in self.elem_list])) == 1
+
+    @property
+    def elem_type(self) -> Type[ElementT]:
+        return type(self.elem_list[0])
+
+    @property
+    def elem_shape(self) -> Tuple[int, ...]:
+        return self.elem_list[0].shape
 
     def _get_has_a_list(self) -> List[ElementT]:
         return self.elem_list
 
-    def append(self, elem: ElementT):
-        if self.elem_type is None:
-            self.elem_type = type(elem)
-        if self.elem_shape is None:
-            self.elem_shape = elem.shape
-        assert type(elem) == self.elem_type
-        assert elem.shape == self.elem_shape
-        self.elem_list.append(elem)
+    def append(self, *args, **kwargs):  # type: ignore
+        raise NotImplementedError("deleted method")
+
+    def dump(self, episode_dir_path: pathlib.Path) -> None:
+        file_path = episode_dir_path / "sequence-{}.npy".format(self.elem_type.__name__)
+        assert issubclass(self.elem_type, PrimitiveElementBase)
+        seq = np.array([e.numpy() for e in self.elem_list])  # type: ignore
+        np.save(file_path, seq)
+
+    @classmethod
+    def load(
+        cls, episode_dir_path: pathlib.Path, elem_type: Type[ElementT]
+    ) -> "ElementSequence[ElementT]":
+        file_path = episode_dir_path / "sequence-{}.npy".format(elem_type.__name__)
+        seq = np.load(file_path)
+        elem_list = [elem_type(arr) for arr in seq]
+        return ElementSequence(elem_list)
+
+    @classmethod
+    def load_all(cls, episode_dir_path: pathlib.Path) -> "Dict[Type[ElementBase], ElementSequence]":
+        d = {}
+        for p in episode_dir_path.iterdir():
+            result = re.match(r"sequence-(\w+).npy", p.name)
+            if result is not None:
+                elem_type = get_element_type(result.group(1))
+                d[elem_type] = cls.load(episode_dir_path, elem_type)  # type: ignore
+        return d
 
 
 def create_composite_image_sequence(
     composite_image_type: Type[CompositeImageT],
-    elem_seqs: List[ElementSequence[PrimitiveImageBase]],
+    elemseq_list: List[ElementSequence[PrimitiveImageBase]],
 ) -> ElementSequence[CompositeImageT]:
+    """Create composite image ElementSequence from different-typed element sequence
+    composite_image_type: e.g. RGBDImage
+    elemseq_list: e.g. [RGBImage, DepthImage]
+    """
 
-    n_len_seq = len(elem_seqs[0])
-    composite_image_seq = ElementSequence[CompositeImageT]([])
+    n_len_seq = len(elemseq_list[0])
+    composite_image_list = []
     for i in range(n_len_seq):
-        composite_image = composite_image_type([seq[i] for seq in elem_seqs])
-        composite_image_seq.append(composite_image)
-    return composite_image_seq
+        composite_image: CompositeImageT = composite_image_type([seq[i] for seq in elemseq_list])
+        composite_image_list.append(composite_image)
+    return ElementSequence(composite_image_list)
 
 
 class TypeShapeTableMixin:
@@ -542,11 +580,23 @@ class TimeStampSequence(HasAList[float]):
         eps = 1e-8
         return [i for i, t in enumerate(self._data) if t > time - eps][0]
 
+    def dump(self, episode_dir_path: pathlib.Path) -> None:
+        file_path = episode_dir_path / "time_stamp_sequence.npy"
+        np.save(file_path, np.array(self._data))
+
+    @classmethod
+    def load(cls, episode_dir_path: pathlib.Path) -> Optional["TimeStampSequence"]:
+        file_path = episode_dir_path / "time_stamp_sequence.npy"
+        if file_path.exists():
+            seq = np.load(file_path).tolist()
+            return TimeStampSequence(seq)
+        else:
+            return None
+
 
 @dataclass(frozen=True)
 class EpisodeData(TypeShapeTableMixin):
     sequence_dict: Dict[Type[ElementBase], ElementSequence[ElementBase]]
-    type_shape_table: Dict[Type[ElementBase], Tuple[int, ...]]
     time_stamp_seq: Optional[TimeStampSequence] = None
     metadata: Optional[MetaData] = None
 
@@ -559,6 +609,13 @@ class EpisodeData(TypeShapeTableMixin):
         assert TerminateFlag in self.sequence_dict
         return len(self.sequence_dict[TerminateFlag])
 
+    @property
+    def type_shape_table(self) -> Dict[Type[ElementBase], Tuple[int, ...]]:
+        dic = {}
+        for key, seq in self.sequence_dict.items():
+            dic[key] = seq.elem_shape
+        return dic
+
     @staticmethod
     def create_default_terminate_flag_seq(n_length) -> ElementSequence[TerminateFlag]:
         flag_lst = [TerminateFlag.from_bool(False) for _ in range(n_length - 1)]
@@ -603,15 +660,12 @@ class EpisodeData(TypeShapeTableMixin):
             sequence_list.append(terminate_flag_seq)
             types.append(TerminateFlag)
 
-        shapes = [seq[0].shape for seq in sequence_list]
-        type_shape_table = dict({t: s for (t, s) in zip(types, shapes)})
-
         n_type = len(set(types))
         all_different_type = n_type == len(sequence_list)
         assert all_different_type, "all sequences must have different type"
 
         sequence_dict = {seq.elem_type: seq for seq in sequence_list}
-        return cls(sequence_dict, type_shape_table, timestamp_seq, metadata)  # type: ignore
+        return cls(sequence_dict, timestamp_seq, metadata)  # type: ignore
 
     def get_sequence_by_type(self, elem_type: Type[ElementT]) -> ElementSequence[ElementT]:
 
@@ -716,14 +770,41 @@ class EpisodeData(TypeShapeTableMixin):
         clip = ImageSequenceClip([e.to_rgb().numpy() for e in seq], fps=fps)
         clip.write_gif(filename, fps=fps)
 
+    def dump(self, episode_dir_path: pathlib.Path) -> None:
+        episode_dir_path = episode_dir_path.expanduser()
+        episode_dir_path.mkdir(exist_ok=True)
+
+        for elem_type, seq in self.sequence_dict.items():
+            seq.dump(episode_dir_path)
+
+        if self.time_stamp_seq is not None:
+            self.time_stamp_seq.dump(episode_dir_path)
+
+        if self.metadata is not None:
+            with open(episode_dir_path / "metadata.json", mode="w") as f:
+                json.dump(self.metadata, f)
+
+    @classmethod
+    def load(cls, episode_dir_path: pathlib.Path) -> "EpisodeData":
+        episode_dir_path = episode_dir_path.expanduser()
+        type_seq_table = ElementSequence.load_all(episode_dir_path)
+        time_stamp_seq = TimeStampSequence.load(episode_dir_path)
+
+        metadata_path = episode_dir_path / "metadata.json"
+        metadata = None
+        if metadata_path.exists():
+            with open(episode_dir_path / "metadata.json", mode="r") as f:
+                metadata = json.load(f)
+        return cls(type_seq_table, time_stamp_seq, metadata)
+
 
 @dataclass(frozen=True)
-class ChunkSpec(TypeShapeTableMixin):
+class BundleSpec(TypeShapeTableMixin):
     n_episode: int
-    n_episode_intact: int
+    n_untouch_episode: int
     n_average: int
     type_shape_table: Dict[Type[ElementBase], Tuple[int, ...]]
-    extra_info: Optional[MetaData] = None
+    meta_data: Optional[MetaData] = None
 
     def to_dict(self) -> Dict:
         d = asdict(self)
@@ -731,121 +812,237 @@ class ChunkSpec(TypeShapeTableMixin):
         return d
 
     @classmethod
-    def from_dict(cls, d: Dict) -> "ChunkSpec":
+    def from_dict(cls, d: Dict) -> "BundleSpec":
         d["type_shape_table"] = {
             get_element_type(k): tuple(v) for k, v in d["type_shape_table"].items()
         }
         return cls(**d)
 
 
-_chunk_cache: Dict[
-    Tuple[str, Optional[str]], "MultiEpisodeChunk"
-] = {}  # used MultiEpisodeChunk.load
+_bundle_cache: Dict[Tuple[str, Optional[str]], "EpisodeBundle"] = {}  # used EpisodeBundle.load
 
 
 @dataclass
-class MultiEpisodeChunk(HasAList[EpisodeData], TypeShapeTableMixin):
-    data_list: List[EpisodeData]
-    data_list_intact: List[EpisodeData]
-    type_shape_table: Dict[Type[ElementBase], Tuple[int, ...]]
-    spec: ChunkSpec
+class EpisodeBundle(HasAList[EpisodeData], TypeShapeTableMixin):
+    """Bundle of episode
+    The collection of episodes.
+    we call it 'bundle' because 'Dataset' is already used by pytorch
+    """
+
+    _episode_list: List[EpisodeData]
+    _untouch_episode_list: List[EpisodeData]
+    _metadata: Optional[MetaData] = None
     _postfix: Optional[str] = None
 
     def _get_has_a_list(self) -> List[EpisodeData]:
-        return self.data_list
+        return self._episode_list
+
+    @property
+    def type_shape_table(self) -> Dict[Type[ElementBase], Tuple[int, ...]]:
+        return self._episode_list[0].type_shape_table
+
+    @property
+    def spec(self) -> BundleSpec:
+        n_average = int(sum([len(data) for data in self._episode_list]) / len(self._episode_list))
+        spec = BundleSpec(
+            len(self._episode_list) - setting.n_untouch_episode,
+            setting.n_untouch_episode,
+            n_average,
+            self.type_shape_table,
+            self._metadata,
+        )
+        return spec
 
     @classmethod
     def from_data_list(
         cls,
         data_list: List[EpisodeData],
-        extra_info: Optional[MetaData] = None,
+        meta_data: Optional[MetaData] = None,
         shuffle: bool = True,
-        with_intact_data: bool = True,
-    ) -> "MultiEpisodeChunk":
+        leave_untouch_episode: bool = True,
+    ) -> "EpisodeBundle":
 
         set_types = set(functools.reduce(operator.add, [list(data.types()) for data in data_list]))
 
         n_type_appeared = len(set_types)
         n_type_expected = len(data_list[0].types())
-        assert_equal_with_message(n_type_appeared, n_type_expected, "num of element in chunk")
+        assert_equal_with_message(n_type_appeared, n_type_expected, "num of element type in bundle")
 
-        data_list_intact = []
-        if with_intact_data:
-            n_data_intact = setting.n_data_intact
-            assert n_data_intact > 0
-            interval = len(data_list) // n_data_intact
-            indices_intact = [interval * i for i in range(n_data_intact)]
+        untouch_episode_list = []
+        if leave_untouch_episode:
+            n_episode_untouch = setting.n_untouch_episode
+            assert n_episode_untouch > 0
+            interval = len(data_list) // n_episode_untouch
+            indices_untouch = [interval * i for i in range(n_episode_untouch)]
             # sorted is necessary because pop changes index
-            for idx in sorted(indices_intact, reverse=True):
-                data_list_intact.append(data_list.pop(idx))
+            for idx in sorted(indices_untouch, reverse=True):
+                untouch_episode_list.append(data_list.pop(idx))
 
         # use fixed random seed where it's scope is only in this function
         # The reason why I used fixed seed is to keep two different shuffled
-        # chunk generated has the contents in the same order.
-        # This is particulary important say you make a chunk with 5hz and another
-        # with 20hz, and do some computation using both of chunks.
+        # bundle generated has the contents in the same order.
+        # This is particulary important say you make a bundl with 5hz and another
+        # with 20hz, and do some computation using both of bundle.
         rn = random.Random()
         rn.seed(0)
         if shuffle:
             rn.shuffle(data_list)
 
-        type_shape_table = data_list[0].type_shape_table
-
-        # chunk spec
-        n_average = int(sum([len(data) for data in data_list]) / len(data_list))
-        spec = ChunkSpec(
-            len(data_list), len(data_list_intact), n_average, type_shape_table, extra_info
-        )
-        return cls(data_list, data_list_intact, type_shape_table, spec)
-
-    @classmethod
-    def load(
-        cls, project_name: Optional[str] = None, postfix: Optional[str] = None
-    ) -> "MultiEpisodeChunk":
-
-        if project_name is None:
-            project_name = setting.primary_project_name
-        assert isinstance(project_name, str)  # for mypy check
-
-        if project_name not in _chunk_cache:
-            _chunk_cache[(project_name, postfix)] = load_object(cls, project_name, postfix)
-        chunk = _chunk_cache[(project_name, postfix)]
-        assert chunk._postfix == postfix
-        return chunk
+        return cls(data_list, untouch_episode_list, meta_data)
 
     @staticmethod
     def spec_file_path(project_name: Optional[str] = None, postfix: Optional[str] = None) -> Path:
 
         project_path = get_project_path(project_name)
         if postfix is None:
-            yaml_file_path = project_path / "chunk_spec.yaml"
+            yaml_file_path = project_path / "bundle_spec.yaml"
         else:
-            yaml_file_path = project_path / "chunk_spec-{}.yaml".format(postfix)
+            yaml_file_path = project_path / "bundle_spec-{}.yaml".format(postfix)
         return yaml_file_path
 
     @classmethod
     def load_spec(
         cls, project_name: Optional[str] = None, postfix: Optional[str] = None
-    ) -> ChunkSpec:
+    ) -> BundleSpec:
 
         yaml_file_path = cls.spec_file_path(project_name, postfix)
         with yaml_file_path.open(mode="r") as f:
             d = yaml.safe_load(f)
-            spec = ChunkSpec.from_dict(d)
+            spec = BundleSpec.from_dict(d)
         return spec
 
-    def dump(self, project_name: Optional[str] = None, postfix: Optional[str] = None) -> None:
+    @classmethod
+    def _load(cls, bundle_dir_path: pathlib.Path, postfix: Optional[str]) -> "EpisodeBundle":
+        def load_episodes(str_startswitdth: str):
+            episode_names: List[str] = natsort.natsorted(
+                [p.name for p in bundle_dir_path.iterdir() if p.name.startswith(str_startswitdth)],
+            )  # type: ignore
+            episode_list = []
+            for episode_name in episode_names:
+                episode_dir_path = bundle_dir_path / episode_name
+                episode_list.append(EpisodeData.load(episode_dir_path))
+            return episode_list
+
+        episode_list = load_episodes("episode")
+        untouch_episode_list = load_episodes("untouch_episode")
+
+        metadata_file_path = bundle_dir_path / "metadata.yaml"
+        with metadata_file_path.open(mode="r") as f:
+            metadata = yaml.safe_load(f)
+        bundle = EpisodeBundle(episode_list, untouch_episode_list, metadata, postfix)
+        return bundle
+
+    @classmethod
+    def load(
+        cls,
+        project_name: Optional[str] = None,
+        postfix: Optional[str] = None,
+        use_tar: bool = True,
+    ) -> "EpisodeBundle":
+        """load bundle
+        use_tar: default True. if True, load tar archive, otherwise load from a directory
+        """
+
+        if project_name is None:
+            project_name = setting.primary_project_name
+        assert isinstance(project_name, str)  # for mypy check
+
+        if (project_name, postfix) not in _bundle_cache:
+
+            bundle_file_without_ext = "EpisodeBundle"
+            if postfix is not None:
+                bundle_file_without_ext += "-{}".format(postfix)
+
+            if use_tar:
+                bundle_tar = bundle_file_without_ext + ".tar"
+                bundle_tar_path = get_project_path(project_name) / bundle_tar
+
+                with tempfile.TemporaryDirectory() as tmp_dir:
+                    tmp_dir_path = pathlib.Path(tmp_dir)
+                    # TODO: use python's tarfile library
+                    subprocess.run(
+                        "cd {} && tar -xf {}".format(tmp_dir_path, bundle_tar_path), shell=True
+                    )
+                    bundle_dir_path = tmp_dir_path / bundle_file_without_ext
+                    bundle = cls._load(bundle_dir_path, postfix)
+
+            else:
+                bundle_dir_path = get_project_path(project_name) / bundle_file_without_ext
+                bundle = cls._load(bundle_dir_path, postfix)
+
+            _bundle_cache[(project_name, postfix)] = bundle
+
+        bundle = _bundle_cache[(project_name, postfix)]
+        return bundle
+
+    def dump(
+        self,
+        project_name: Optional[str] = None,
+        postfix: Optional[str] = None,
+        use_tar: bool = True,
+    ) -> None:
+        """dump the bundle
+        use_tar: defalut True. if True, save as tar archive, otherwise save as a directory
+
+        NOTE: tar is great because it's immutable and no trouble when downloading from gdrive
+        and it can be easily viewed on common file viewer.
+
+        TODO: Adding option for compression rate control is a future option.
+        Please send me a PR. Currently, no compreesion is applied.
+        """
+
         self._postfix = postfix
-        dump_object(self, project_name, postfix)
-        yaml_file_path = self.spec_file_path(project_name, postfix)
-        with yaml_file_path.open(mode="w") as f:
+        bundle_file_without_ext = "EpisodeBundle" + (
+            "" if postfix is None else "-{}".format(postfix)
+        )
+
+        with tempfile.TemporaryDirectory() as tmp_dir:
+            tmp_dir_path = pathlib.Path(tmp_dir)
+
+            bundle_dir_path = tmp_dir_path / bundle_file_without_ext
+            bundle_dir_path.mkdir(parents=True, exist_ok=True)
+
+            for i, episode in enumerate(self._episode_list):
+                episode_dir_path = bundle_dir_path / "episode{}".format(i)
+                episode.dump(episode_dir_path)
+
+            for i, episode in enumerate(self._untouch_episode_list):
+                episode_dir_path = bundle_dir_path / "untouch_episode{}".format(i)
+                episode.dump(episode_dir_path)
+
+            metadata_file_path = bundle_dir_path / "metadata.yaml"
+            with metadata_file_path.open(mode="w") as f:
+                yaml.dump(self._metadata, f, default_flow_style=False, sort_keys=False)
+
+            if use_tar:
+                tarfile = bundle_file_without_ext + ".tar"
+                tarfile_full = get_project_path(project_name) / tarfile
+                if tarfile_full.exists():
+                    os.remove(tarfile_full)
+
+                # TODO: using python tarfile is clean appoach. If get annoyed, please send a PR
+                cmd = "cd {} && tar cvf {} *".format(tmp_dir_path, tarfile_full)
+                subprocess.run(cmd, shell=True)
+            else:  # just move things to project directory
+                project_path = get_project_path(project_name)
+                destination_path = project_path / bundle_file_without_ext
+                if destination_path.exists():
+                    shutil.rmtree(destination_path)
+                # https://bugs.python.org/issue34069
+                shutil.move(str(bundle_dir_path), str(destination_path))
+
+        # extra dump just for debugging (the following info is not requried to load bundle)
+        spec_file_path = self.spec_file_path(project_name, postfix)
+        with spec_file_path.open(mode="w") as f:
             yaml.dump(self.spec.to_dict(), f, default_flow_style=False, sort_keys=False)
 
-    def get_intact_chunk(self) -> "MultiEpisodeChunk":
-        return MultiEpisodeChunk(self.data_list_intact, [], self.type_shape_table, self.spec)
+    def get_untouch_bundle(self) -> "EpisodeBundle":
+        """get episode bundle which is not used for training."""
+        return EpisodeBundle(self._untouch_episode_list, [], self._metadata, self._postfix)
 
-    def get_not_intact_chunk(self) -> "MultiEpisodeChunk":
-        return MultiEpisodeChunk(self.data_list, [], self.type_shape_table, self.spec)
+    def get_touch_bundle(self) -> "EpisodeBundle":
+        """get episode bundle which is used for training"""
+        return EpisodeBundle(self._episode_list, [], self._metadata, self._postfix)
 
     def plot_vector_histories(
         self,
@@ -863,7 +1060,7 @@ class MultiEpisodeChunk(HasAList[EpisodeData], TypeShapeTableMixin):
 
             y_min, y_max = +np.inf, -np.inf  # will be updated in the following loop
 
-            for data in self.data_list:
+            for data in self._episode_list:
                 seq = data.get_sequence_by_type(AngleVector)
                 single_seq = np.array([e.numpy()[i_dim] for e in seq])
                 y_min = min(y_min, np.min(single_seq))
